@@ -3,6 +3,10 @@
 
 #include <orbit/GIOP/giop-types.h>
 
+G_BEGIN_DECLS
+
+#ifdef ORBIT2_INTERNAL_API
+
 /* This is also defined in IIOP-types.c */
 void giop_byteswap(guchar *outdata,
 		   const guchar *data,
@@ -20,4 +24,8 @@ G_INLINE_FUNC void giop_byteswap(guchar *outdata,
 }
 #endif
 
+#endif /* ORBIT2_INTERNAL_API */
+
+G_END_DECLS
+
 #endif
